import { __values, __assign, __awaiter, __generator, __read } from 'tslib';
import { inject } from '@angular/core';
import { AngularFirestore } from '@angular/fire/firestore';
import { runStoreAction, StoreActions, withTransaction, arrayUpdate, arrayRemove, arrayUpsert, arrayAdd } from '@datorama/akita';
import 'firebase/firestore';
import { isObservable, of, combineLatest, Subject, race, throwError } from 'rxjs';
import { switchMap, tap, map, takeUntil, distinctUntilChanged, filter, share, finalize } from 'rxjs/operators';
import { Router, UrlTree } from '@angular/router';
import { AngularFireAuth } from '@angular/fire/auth';
import { auth } from 'firebase/app';

// Helper to retrieve the id and path of a document in the collection
function getIdAndPath(options, collectionPath) {
    var path = '';
    var id = '';
    if (options['id']) {
        if (!collectionPath) {
            throw new Error('You should provide the colletion path with the id');
        }
        id = options['id'];
        path = collectionPath + "/" + id;
    }
    else if (options['path']) {
        path = options['path'];
        var part = path.split('/');
        if (part.length % 2 !== 0) {
            throw new Error("Path " + path + " doesn't look like a Firestore's document path");
        }
        id = part[part.length - 1];
    }
    else {
        throw new Error("You should provide either an \"id\" OR a \"path\".");
    }
    return { id: id, path: path };
}

/** Set the loading parameter of a specific store */
function setLoading(storeName, loading) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { loading: loading }
        }
    });
}
/** Reset the store to an empty array */
function resetStore(storeName) {
    runStoreAction(storeName, StoreActions.SetEntities, {
        payload: {
            data: []
        }
    });
}
/**  */
function setActive(storeName, active) {
    runStoreAction(storeName, StoreActions.Update, {
        payload: {
            data: { active: active }
        }
    });
}
/** Create or update one or several entities in the store */
function upsertStoreEntity(storeName, data) {
    var payload = { data: data };
    runStoreAction(storeName, StoreActions.UpsertEntities, { payload: payload });
}
/** Remove one or several entities in the store */
function removeStoreEntity(storeName, entityIds) {
    var payload = { entityIds: entityIds };
    runStoreAction(storeName, StoreActions.RemoveEntities, { payload: payload });
}
/** Update one or several entities in the store */
function updateStoreEntity(storeName, entityIds, data) {
    var payload = {
        data: data,
        entityIds: entityIds
    };
    runStoreAction(storeName, StoreActions.UpdateEntities, { payload: payload });
}
/** Sync a specific store with actions from Firestore */
function syncStoreFromDocAction(storeName, actions, idKey, formatFromFirestore) {
    var e_1, _a, _b;
    if (idKey === void 0) { idKey = 'id'; }
    setLoading(storeName, false);
    if (actions.length === 0) {
        return;
    }
    try {
        for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
            var action = actions_1_1.value;
            var id = action.payload.doc.id;
            var entity = formatFromFirestore(action.payload.doc.data());
            switch (action.type) {
                case 'added': {
                    upsertStoreEntity(storeName, __assign((_b = {}, _b[idKey] = id, _b), entity));
                    break;
                }
                case 'removed': {
                    removeStoreEntity(storeName, id);
                    break;
                }
                case 'modified': {
                    updateStoreEntity(storeName, id, entity);
                    break;
                }
            }
        }
    }
    catch (e_1_1) { e_1 = { error: e_1_1 }; }
    finally {
        try {
            if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
        }
        finally { if (e_1) throw e_1.error; }
    }
}
/** Sync a specific store with actions from Firestore */
function syncStoreFromDocActionSnapshot(storeName, action, idKey, formatFromFirestore) {
    var _a;
    if (idKey === void 0) { idKey = 'id'; }
    setLoading(storeName, false);
    var id = action.payload.id;
    var entity = formatFromFirestore(action.payload.data());
    if (!action.payload.exists) {
        removeStoreEntity(storeName, id);
    }
    else {
        upsertStoreEntity(storeName, __assign((_a = {}, _a[idKey] = id, _a), entity));
    }
}

/**
 * Get the store name of a store to be synced
 */
function getStoreName(store, storeOptions) {
    if (storeOptions === void 0) { storeOptions = {}; }
    if (!store && !storeOptions.storeName) {
        throw new Error('You should either provide a store name or inject a store instance in constructor');
    }
    return storeOptions.storeName || store.storeName;
}

/** Get the params from a path */
function getPathParams(path) {
    return path.split('/')
        .filter(function (segment) { return segment.charAt(0) === ':'; })
        .map(function (segment) { return segment.substr(1); });
}
/**
 * Transform a path based on the params
 * @param path The path with params starting with "/:"
 * @param params A map of id params
 * @example pathWithParams('movies/:movieId/stakeholder/:shId', { movieId, shId })
 */
function pathWithParams(path, params) {
    return path.split('/').map(function (segment) {
        if (segment.charAt(0) === ':') {
            var key = segment.substr(1);
            if (!params[key]) {
                throw new Error("Required parameter " + key + " from " + path + " doesn't exist in params " + JSON.stringify(params));
            }
            return params[key];
        }
        else {
            return segment;
        }
    }).join('/');
}

/**
 * Same as Object.getOwnPropertyDescriptor, but recursively checks prototype chain excluding passed currentClass
 * @param instance Instance to check on
 * @param params Property name to check
 * @param currentClass Checks prototype chain until this class
 * @example getPropertyDescriptor(this, 'path', CollectionService)
 */
function getPropertyDescriptor(instance, property, currentClass) {
    if (currentClass === void 0) { currentClass = Object; }
    var prototype = Object.getPrototypeOf(instance);
    if (!prototype || !(prototype instanceof currentClass))
        return;
    return Object.getOwnPropertyDescriptor(prototype, property) || getPropertyDescriptor(prototype, property, currentClass);
}
/**
 * Check prototype chain for a specific getter function, excluding parent class
 * @param instance Instance of the class to check on
 * @param parentClass Parent class of the instance
 * @param property Property name to check
 * @example hasChildGetter(this, CollectionService, 'path')
 */
function hasChildGetter(instance, parentClass, property) {
    var descriptor = getPropertyDescriptor(instance, property, parentClass);
    return descriptor && descriptor.get && true;
}

function isArray(entityOrArray) {
    return Array.isArray(entityOrArray);
}
/** check is an Atomic write is a transaction */
function isTransaction(write) {
    return write && !!write['get'];
}
var CollectionService = /** @class */ (function () {
    function CollectionService(store, collectionPath, db) {
        this.store = store;
        this.collectionPath = collectionPath;
        // keep memory of the current ids to listen to (for syncManyDocs)
        this.idsToListen = {};
        if (!hasChildGetter(this, CollectionService, 'path') && !this.constructor['path'] && !this.collectionPath) {
            throw new Error('You should provide a path to the collection');
        }
        try {
            this.db = db || inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionService requires AngularFirestore.');
        }
    }
    CollectionService.prototype.getPath = function (options) {
        return (options && options.params)
            ? pathWithParams(this.path, options.params)
            : this.currentPath;
    };
    Object.defineProperty(CollectionService.prototype, "idKey", {
        get: function () {
            return this.constructor['idKey']
                || this.store ? this.store.idKey : 'id';
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "path", {
        /** The path to the collection in Firestore */
        get: function () {
            return this.constructor['path'] || this.collectionPath;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "currentPath", {
        /** A snapshot of the path */
        get: function () {
            if (isObservable(this.path)) {
                throw new Error('Cannot get a snapshot of the path if it is an Observable');
            }
            return this.path;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionService.prototype, "collection", {
        /**
         * The Angular Fire collection
         * @notice If path is an observable, it becomes an observable.
         */
        get: function () {
            return this.db.collection(this.currentPath);
        },
        enumerable: true,
        configurable: true
    });
    /** @deprecated Please use @see formatToFirestore */
    CollectionService.prototype.preFormat = function (document) {
        return document;
    };
    /**
     * Function triggered when adding/updating data to firestore
     * @note should be overrided
     */
    CollectionService.prototype.formatToFirestore = function (entity) {
        return entity;
    };
    /**
     * Function triggered when getting data from firestore
     * @note should be overrided
     */
    CollectionService.prototype.formatFromFirestore = function (entity) {
        return entity;
    };
    Object.defineProperty(CollectionService.prototype, "config", {
        /** The config given by the `CollectonConfig` */
        get: function () {
            return {
                path: this.constructor['path'],
                idKey: this.constructor['idKey']
            };
        },
        enumerable: true,
        configurable: true
    });
    CollectionService.prototype.syncCollection = function (pathOrQuery, queryOrOptions, syncOptions) {
        var _this = this;
        if (pathOrQuery === void 0) { pathOrQuery = this.currentPath; }
        if (syncOptions === void 0) { syncOptions = { loading: true }; }
        var path;
        var queryFn;
        // check type of pathOrQuery
        if (typeof pathOrQuery === 'function') {
            queryFn = pathOrQuery;
            path = this.getPath(queryOrOptions);
        }
        else if (typeof pathOrQuery === 'object') {
            syncOptions = pathOrQuery;
            path = this.getPath(syncOptions);
        }
        else if (typeof pathOrQuery === 'string') {
            path = pathOrQuery;
        }
        else {
            path = this.getPath(syncOptions);
        }
        // check type of queryOrOptions
        if (typeof queryOrOptions === 'function') {
            queryFn = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            syncOptions = queryOrOptions;
        }
        var storeName = getStoreName(this.store, syncOptions);
        // reset has to happen before setLoading, otherwise it will also reset the loading state
        if (syncOptions.reset) {
            resetStore(storeName);
        }
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        // Start Listening
        return this.db.collection(path, queryFn).stateChanges().pipe(withTransaction(function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey, function (entity) { return _this.formatFromFirestore(entity); }); }));
    };
    CollectionService.prototype.syncCollectionGroup = function (idOrQuery, queryOrOption, syncOptions) {
        var _this = this;
        if (idOrQuery === void 0) { idOrQuery = this.currentPath; }
        if (syncOptions === void 0) { syncOptions = { loading: true }; }
        var path;
        var query;
        if (typeof idOrQuery === 'string') {
            path = idOrQuery;
        }
        else if (typeof idOrQuery === 'function') {
            path = this.currentPath;
            query = idOrQuery;
        }
        else if (typeof idOrQuery === 'object') {
            path = this.currentPath;
            syncOptions = idOrQuery;
        }
        else {
            throw new Error('1ier parameter if either a string, a queryFn or a StoreOption');
        }
        if (typeof queryOrOption === 'function') {
            query = queryOrOption;
        }
        else if (typeof queryOrOption === 'object') {
            syncOptions = queryOrOption;
        }
        var storeName = getStoreName(this.store, syncOptions);
        // reset has to happen before setLoading, otherwise it will also reset the loading state
        if (syncOptions.reset) {
            resetStore(storeName);
        }
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        var collectionId = path.split('/').pop();
        return this.db.collectionGroup(collectionId, query).stateChanges().pipe(withTransaction(function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey, function (entity) { return _this.formatFromFirestore(entity); }); }));
    };
    CollectionService.prototype.syncManyDocs = function (ids$, syncOptions) {
        var _this = this;
        if (syncOptions === void 0) { syncOptions = { loading: true }; }
        if (!isObservable(ids$)) {
            ids$ = of(ids$);
        }
        var storeName = getStoreName(this.store, syncOptions);
        // reset has to happen before setLoading, otherwise it will also reset the loading state
        if (syncOptions.reset) {
            resetStore(storeName);
        }
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        return ids$.pipe(switchMap((function (ids) {
            // Remove previous ids that have changed
            var previousIds = _this.idsToListen[storeName];
            if (previousIds) {
                var idsToRemove = previousIds.filter(function (id) { return !ids.includes(id); });
                removeStoreEntity(storeName, idsToRemove);
            }
            _this.idsToListen[storeName] = ids;
            // Return empty array if no ids are provided
            if (!ids.length) {
                return of([]);
            }
            // Sync all docs
            var syncs = ids.map(function (id) {
                var path = _this.getPath(syncOptions) + "/" + id;
                return _this.db.doc(path).snapshotChanges();
            });
            return combineLatest(syncs).pipe(tap(function (actions) { return actions.map(function (action) {
                syncStoreFromDocActionSnapshot(storeName, action, _this.idKey, function (entity) { return _this.formatFromFirestore(entity); });
            }); }));
        })));
    };
    /**
     * Stay in sync with one document
     * @param docOptions An object with EITHER `id` OR `path`.
     * @note We need to use id and path because there is no way to differentiate them.
     * @param syncOptions Options on the store to sync to
     */
    CollectionService.prototype.syncDoc = function (docOptions, syncOptions) {
        var _this = this;
        if (syncOptions === void 0) { syncOptions = { loading: false }; }
        var storeName = getStoreName(this.store, syncOptions);
        var collectionPath = this.getPath(syncOptions);
        var _a = getIdAndPath(docOptions, collectionPath), id = _a.id, path = _a.path;
        // reset has to happen before setLoading, otherwise it will also reset the loading state
        if (syncOptions.reset) {
            resetStore(storeName);
        }
        if (syncOptions.loading) {
            setLoading(storeName, true);
        }
        return this.db.doc(path).valueChanges().pipe(map(function (entity) {
            var _a;
            if (!entity) {
                setLoading(storeName, false);
                // note: We don't removeEntity as it would result in weird behavior
                return undefined;
            }
            var data = _this.formatFromFirestore(__assign((_a = {}, _a[_this.idKey] = id, _a), entity));
            upsertStoreEntity(storeName, data);
            setLoading(storeName, false);
            return data;
        }));
    };
    CollectionService.prototype.syncActive = function (options, syncOptions) {
        var _this = this;
        var storeName = getStoreName(this.store, syncOptions);
        if (Array.isArray(options)) {
            return this.syncManyDocs(options, syncOptions).pipe(tap(function (_) { return setActive(storeName, options); }));
        }
        else {
            return this.syncDoc(options, syncOptions).pipe(tap(function (entity) { return entity ? setActive(storeName, entity[_this.idKey]) : null; }));
        }
    };
    CollectionService.prototype.getRef = function (idOrQuery, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        var path = this.getPath(options);
        // If path targets a collection ( odd number of segments after the split )
        if (typeof idOrQuery === 'string') {
            return this.db.doc(path + "/" + idOrQuery).ref;
        }
        if (Array.isArray(idOrQuery)) {
            return idOrQuery.map(function (id) { return _this.db.doc(path + "/" + id).ref; });
        }
        else if (typeof idOrQuery === 'object') {
            var subpath = this.getPath(idOrQuery);
            return this.db.collection(subpath).ref;
        }
        else {
            return this.db.collection(path, idOrQuery).ref;
        }
    };
    CollectionService.prototype.getValue = function (idOrQuery, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var path, snapshot, docs, ref, snaphot, subpath, snapshot, snapshot;
            var _a;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        path = this.getPath(options);
                        if (!(typeof idOrQuery === 'string')) return [3 /*break*/, 2];
                        return [4 /*yield*/, this.db.doc(path + "/" + idOrQuery).ref.get()];
                    case 1:
                        snapshot = _b.sent();
                        return [2 /*return*/, snapshot.exists
                                ? this.formatFromFirestore(__assign(__assign({}, snapshot.data()), (_a = {}, _a[this.idKey] = snapshot.id, _a)))
                                : null];
                    case 2:
                        if (!Array.isArray(idOrQuery)) return [3 /*break*/, 4];
                        return [4 /*yield*/, Promise.all(idOrQuery.map(function (id) {
                                return _this.db.doc(path + "/" + id).ref.get();
                            }))];
                    case 3:
                        docs = _b.sent();
                        return [3 /*break*/, 10];
                    case 4:
                        if (!(typeof idOrQuery === 'function')) return [3 /*break*/, 6];
                        ref = this.db.collection(path).ref;
                        return [4 /*yield*/, idOrQuery(ref).get()];
                    case 5:
                        snaphot = _b.sent();
                        docs = snaphot.docs;
                        return [3 /*break*/, 10];
                    case 6:
                        if (!(typeof idOrQuery === 'object')) return [3 /*break*/, 8];
                        subpath = this.getPath(idOrQuery);
                        return [4 /*yield*/, this.db.collection(subpath).ref.get()];
                    case 7:
                        snapshot = _b.sent();
                        docs = snapshot.docs;
                        return [3 /*break*/, 10];
                    case 8: return [4 /*yield*/, this.db.collection(path, idOrQuery).ref.get()];
                    case 9:
                        snapshot = _b.sent();
                        docs = snapshot.docs;
                        _b.label = 10;
                    case 10: return [2 /*return*/, docs.filter(function (doc) { return doc.exists; })
                            .map(function (doc) {
                            var _a;
                            return _this.formatFromFirestore(__assign(__assign({}, doc.data()), (_a = {}, _a[_this.idKey] = doc.id, _a)));
                        })];
                }
            });
        });
    };
    CollectionService.prototype.valueChanges = function (idOrQuery, options) {
        var _this = this;
        if (options === void 0) { options = {}; }
        var path = this.getPath(options);
        // If path targets a collection ( odd number of segments after the split )
        if (typeof idOrQuery === 'string') {
            return this.db.doc(path + "/" + idOrQuery).valueChanges().pipe(map(function (doc) { return _this.formatFromFirestore(doc); }));
        }
        var docs$;
        if (Array.isArray(idOrQuery)) {
            docs$ = combineLatest(idOrQuery.map(function (id) { return _this.db.doc(path + "/" + id).valueChanges(); }));
        }
        else if (typeof idOrQuery === 'function') {
            docs$ = this.db.collection(path, idOrQuery).valueChanges();
        }
        else if (typeof idOrQuery === 'object') {
            var subpath = this.getPath(idOrQuery);
            docs$ = this.db.collection(subpath).valueChanges();
        }
        else {
            docs$ = this.db.collection(path, idOrQuery).valueChanges();
        }
        return docs$.pipe(map(function (docs) { return docs.map(function (doc) { return _this.formatFromFirestore(doc); }); }));
    };
    ///////////
    // WRITE //
    ///////////
    /**
     * Create a batch object.
     * @note alias for `angularFirestore.firestore.batch()`
     */
    CollectionService.prototype.batch = function () {
        return this.db.firestore.batch();
    };
    /**
     * Run a transaction
     * @note alias for `angularFirestore.firestore.runTransaction()`
     */
    CollectionService.prototype.runTransaction = function (cb) {
        return this.db.firestore.runTransaction(function (tx) { return cb(tx); });
    };
    /**
     * Create or update documents
     * @param documents One or many documents
     * @param options options to write the document on firestore
     */
    CollectionService.prototype.upsert = function (documents, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var doesExist, toAdd, toUpdate, documents_1, documents_1_1, doc, e_1_1;
            var e_1, _a;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        doesExist = function (doc) { return __awaiter(_this, void 0, void 0, function () {
                            var id, ref, exists;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        id = doc[this.idKey];
                                        if (!id) return [3 /*break*/, 1];
                                        return [2 /*return*/, true];
                                    case 1:
                                        ref = this.getRef(doc[this.idKey]);
                                        return [4 /*yield*/, (isTransaction(options.write) ? options.write.get(ref) : ref.get())];
                                    case 2:
                                        exists = (_a.sent()).exists;
                                        return [2 /*return*/, exists];
                                }
                            });
                        }); };
                        if (!!isArray(documents)) return [3 /*break*/, 2];
                        return [4 /*yield*/, doesExist(documents)];
                    case 1: return [2 /*return*/, (_b.sent())
                            ? this.update(documents, options).then(function (_) { return documents[_this.idKey]; })
                            : this.add(documents, options)];
                    case 2:
                        toAdd = [];
                        toUpdate = [];
                        _b.label = 3;
                    case 3:
                        _b.trys.push([3, 8, 9, 10]);
                        documents_1 = __values(documents), documents_1_1 = documents_1.next();
                        _b.label = 4;
                    case 4:
                        if (!!documents_1_1.done) return [3 /*break*/, 7];
                        doc = documents_1_1.value;
                        return [4 /*yield*/, doesExist(doc)];
                    case 5:
                        (_b.sent())
                            ? toUpdate.push(doc)
                            : toAdd.push(doc);
                        _b.label = 6;
                    case 6:
                        documents_1_1 = documents_1.next();
                        return [3 /*break*/, 4];
                    case 7: return [3 /*break*/, 10];
                    case 8:
                        e_1_1 = _b.sent();
                        e_1 = { error: e_1_1 };
                        return [3 /*break*/, 10];
                    case 9:
                        try {
                            if (documents_1_1 && !documents_1_1.done && (_a = documents_1.return)) _a.call(documents_1);
                        }
                        finally { if (e_1) throw e_1.error; }
                        return [7 /*endfinally*/];
                    case 10: return [2 /*return*/, Promise.all([
                            this.add(toAdd, options),
                            this.update(toUpdate, options).then(function (_) { return toUpdate.map(function (doc) { return doc[_this.idKey]; }); })
                        ]).then(function (_a) {
                            var _b = __read(_a, 2), added = _b[0], updated = _b[1];
                            return added.concat(updated);
                        })];
                }
            });
        });
    };
    /**
     * Add a document or a list of document to Firestore
     * @param docs A document or a list of document
     * @param options options to write the document on firestore
     */
    CollectionService.prototype.add = function (documents, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var docs, _a, write, ctx, path, operations, ids;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        docs = (Array.isArray(documents) ? documents : [documents]);
                        _a = options.write, write = _a === void 0 ? this.batch() : _a, ctx = options.ctx;
                        path = this.getPath(options);
                        operations = docs.map(function (doc) { return __awaiter(_this, void 0, void 0, function () {
                            var id, data, ref;
                            var _a;
                            return __generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0:
                                        id = doc[this.idKey] || this.db.createId();
                                        data = this.preFormat(__assign(__assign({}, doc), (_a = {}, _a[this.idKey] = id, _a)));
                                        ref = this.db.doc(path + "/" + id).ref;
                                        write.set(ref, this.formatToFirestore((data)));
                                        if (!this.onCreate) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onCreate(data, { write: write, ctx: ctx })];
                                    case 1:
                                        _b.sent();
                                        _b.label = 2;
                                    case 2: return [2 /*return*/, id];
                                }
                            });
                        }); });
                        return [4 /*yield*/, Promise.all(operations)];
                    case 1:
                        ids = _b.sent();
                        if (!!options.write) return [3 /*break*/, 3];
                        return [4 /*yield*/, write.commit()];
                    case 2:
                        _b.sent();
                        _b.label = 3;
                    case 3: return [2 /*return*/, Array.isArray(documents) ? ids : ids[0]];
                }
            });
        });
    };
    /**
     * Remove one or several document from Firestore
     * @param id A unique or list of id representing the document
     * @param options options to write the document on firestore
     */
    CollectionService.prototype.remove = function (id, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var _a, write, ctx, path, ids, operations;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        _a = options.write, write = _a === void 0 ? this.batch() : _a, ctx = options.ctx;
                        path = this.getPath(options);
                        ids = Array.isArray(id) ? id : [id];
                        operations = ids.map(function (docId) { return __awaiter(_this, void 0, void 0, function () {
                            var ref;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        ref = this.db.doc(path + "/" + docId).ref;
                                        write.delete(ref);
                                        if (!this.onDelete) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onDelete(docId, { write: write, ctx: ctx })];
                                    case 1:
                                        _a.sent();
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        }); });
                        return [4 /*yield*/, Promise.all(operations)];
                    case 1:
                        _b.sent();
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, write.commit()];
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    /** Remove all document of the collection */
    CollectionService.prototype.removeAll = function (options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var path, snapshot, ids;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        path = this.getPath(options);
                        return [4 /*yield*/, this.db.collection(path).ref.get()];
                    case 1:
                        snapshot = _a.sent();
                        ids = snapshot.docs.map(function (doc) { return doc.id; });
                        return [2 /*return*/, this.remove(ids, options)];
                }
            });
        });
    };
    CollectionService.prototype.update = function (idsOrEntity, stateFnOrWrite, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var ids, stateFunction, getData, isEntity, isEntityArray, entityMap_1, ctx, path, _a, write_1, operations;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0:
                        ids = [];
                        isEntity = function (value) {
                            return typeof value === 'object' && value[_this.idKey];
                        };
                        isEntityArray = function (values) {
                            return Array.isArray(values) && values.every(function (value) { return isEntity(value); });
                        };
                        if (isEntity(idsOrEntity)) {
                            ids = [idsOrEntity[this.idKey]];
                            getData = function () { return idsOrEntity; };
                            options = stateFnOrWrite || {};
                        }
                        else if (isEntityArray(idsOrEntity)) {
                            entityMap_1 = new Map(idsOrEntity.map(function (entity) { return [entity[_this.idKey], entity]; }));
                            ids = Array.from(entityMap_1.keys());
                            getData = function (docId) { return entityMap_1.get(docId); };
                            options = stateFnOrWrite || {};
                        }
                        else if (typeof stateFnOrWrite === 'function') {
                            ids = Array.isArray(idsOrEntity) ? idsOrEntity : [idsOrEntity];
                            stateFunction = stateFnOrWrite;
                        }
                        else if (typeof stateFnOrWrite === 'object') {
                            ids = Array.isArray(idsOrEntity) ? idsOrEntity : [idsOrEntity];
                            getData = function () { return stateFnOrWrite; };
                        }
                        else {
                            throw new Error('Passed parameters match none of the function signatures.');
                        }
                        ctx = options.ctx;
                        path = this.getPath(options);
                        if (!Array.isArray(ids) || !ids.length) {
                            return [2 /*return*/];
                        }
                        if (!stateFunction) return [3 /*break*/, 1];
                        return [2 /*return*/, this.db.firestore.runTransaction(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                                var operations;
                                var _this = this;
                                return __generator(this, function (_a) {
                                    operations = ids.map(function (id) { return __awaiter(_this, void 0, void 0, function () {
                                        var ref, snapshot, doc, data;
                                        var _a;
                                        return __generator(this, function (_b) {
                                            switch (_b.label) {
                                                case 0:
                                                    ref = this.db.doc(path + "/" + id).ref;
                                                    return [4 /*yield*/, tx.get(ref)];
                                                case 1:
                                                    snapshot = _b.sent();
                                                    doc = Object.freeze(__assign(__assign({}, snapshot.data()), (_a = {}, _a[this.idKey] = id, _a)));
                                                    data = stateFunction(this.preFormat(doc), tx);
                                                    tx.update(ref, this.formatToFirestore(data));
                                                    if (!this.onUpdate) return [3 /*break*/, 3];
                                                    return [4 /*yield*/, this.onUpdate(data, { write: tx, ctx: ctx })];
                                                case 2:
                                                    _b.sent();
                                                    _b.label = 3;
                                                case 3: return [2 /*return*/, tx];
                                            }
                                        });
                                    }); });
                                    return [2 /*return*/, Promise.all(operations)];
                                });
                            }); })];
                    case 1:
                        _a = options.write, write_1 = _a === void 0 ? this.batch() : _a;
                        operations = ids.map(function (docId) { return __awaiter(_this, void 0, void 0, function () {
                            var doc, data, ref;
                            return __generator(this, function (_a) {
                                switch (_a.label) {
                                    case 0:
                                        doc = Object.freeze(getData(docId));
                                        data = this.preFormat(doc);
                                        if (!docId) {
                                            throw new Error("Document should have an unique id to be updated, but none was found in " + doc);
                                        }
                                        ref = this.db.doc(path + "/" + docId).ref;
                                        write_1.update(ref, this.formatToFirestore(data));
                                        if (!this.onUpdate) return [3 /*break*/, 2];
                                        return [4 /*yield*/, this.onUpdate(data, { write: write_1, ctx: ctx })];
                                    case 1:
                                        _a.sent();
                                        _a.label = 2;
                                    case 2: return [2 /*return*/];
                                }
                            });
                        }); });
                        return [4 /*yield*/, Promise.all(operations)];
                    case 2:
                        _b.sent();
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, write_1.commit()];
                        }
                        return [2 /*return*/];
                }
            });
        });
    };
    return CollectionService;
}());

/** Set the configuration for the collection service */
function CollectionConfig(options) {
    if (options === void 0) { options = {}; }
    return function (constructor) {
        Object.keys(options).forEach(function (key) { return constructor[key] = options[key]; });
    };
}

var initialAuthState = {
    uid: null,
    emailVerified: undefined,
    profile: null,
    loading: false
};

var authProviders = ['github', 'google', 'microsoft', 'facebook', 'twitter', 'email', 'apple'];
/** Verify if provider is part of the list of Authentication provider provided by Firebase Auth */
function isFireAuthProvider(provider) {
    return typeof provider === 'string' && authProviders.includes(provider);
}
/**
 * Get the custom claims of a user. If no key is provided, return the whole claims object
 * @param user The user object returned by Firebase Auth
 * @param roles Keys of the custom claims inside the claim objet
 */
function getCustomClaims(user, roles) {
    return __awaiter(this, void 0, void 0, function () {
        var claims, keys;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, user.getIdTokenResult()];
                case 1:
                    claims = (_a.sent()).claims;
                    if (!roles) {
                        return [2 /*return*/, claims];
                    }
                    keys = Array.isArray(roles) ? roles : [roles];
                    return [2 /*return*/, Object.keys(claims)
                            .filter(function (key) { return keys.includes(key); })
                            .reduce(function (acc, key) {
                            acc[key] = claims[key];
                            return acc;
                        }, {})];
            }
        });
    });
}
/**
 * Get the Authentication Provider based on its name
 * @param provider string literal representing the name of the provider
 */
function getAuthProvider(provider) {
    switch (provider) {
        case 'email': return new auth.EmailAuthProvider();
        case 'facebook': return new auth.FacebookAuthProvider();
        case 'github': return new auth.GithubAuthProvider();
        case 'google': return new auth.GoogleAuthProvider();
        case 'microsoft': return new auth.OAuthProvider('microsoft.com');
        case 'twitter': return new auth.TwitterAuthProvider();
        case 'apple': return new auth.OAuthProvider('apple');
    }
}
var FireAuthService = /** @class */ (function () {
    function FireAuthService(store, db, auth) {
        this.store = store;
        this.collectionPath = 'users';
        this.db = db || inject(AngularFirestore);
        this.auth = auth || inject(AngularFireAuth);
        this.collection = this.db.collection(this.path);
    }
    /**
     * Select the profile in the Firestore
     * @note can be override to point to a different place
     */
    FireAuthService.prototype.selectProfile = function (user) {
        return this.collection.doc(user.uid).valueChanges();
    };
    /**
     * Select the roles for this user. Can be in custom claims or in a Firestore collection
     * @param user The user given by FireAuth
     * @see getCustomClaims to get the custom claims out of the user
     * @note Can be overrided
     */
    FireAuthService.prototype.selectRoles = function (user) {
        return of(null);
    };
    /**
     * Function triggered when getting data from firestore
     * @note should be overrided
     */
    FireAuthService.prototype.formatFromFirestore = function (user) {
        return user;
    };
    /**
     * Function triggered when adding/updating data to firestore
     * @note should be overrided
     */
    FireAuthService.prototype.formatToFirestore = function (user) {
        return user;
    };
    /**
     * Function triggered when transforming a user into a profile
     * @param user The user object from FireAuth
     * @param ctx The context given on signup
     * @note Should be override
     */
    FireAuthService.prototype.createProfile = function (user, ctx) {
        return {
            photoURL: user.photoURL,
            displayName: user.displayName,
        };
    };
    Object.defineProperty(FireAuthService.prototype, "fireAuth", {
        /** @deprecated use 'auth' instead */
        get: function () {
            return this.auth;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FireAuthService.prototype, "user", {
        /**
         * The current sign-in user (or null)
         * @returns a Promise in v6.*.* & a snapshot in v5.*.*
         */
        get: function () {
            return this.auth.currentUser;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FireAuthService.prototype, "idKey", {
        get: function () {
            return this.constructor['idKey'] || 'id';
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(FireAuthService.prototype, "path", {
        /** The path to the profile in firestore */
        get: function () {
            return this.constructor['path'] || this.collectionPath;
        },
        enumerable: true,
        configurable: true
    });
    /** Start listening on User */
    FireAuthService.prototype.sync = function () {
        var _this = this;
        return this.auth.authState.pipe(switchMap(function (user) { return user ? combineLatest([
            of(user),
            _this.selectProfile(user),
            _this.selectRoles(user),
        ]) : of([undefined, undefined, undefined]); }), tap(function (_a) {
            var _b = __read(_a, 3), _c = _b[0], user = _c === void 0 ? {} : _c, userProfile = _b[1], roles = _b[2];
            var profile = _this.formatFromFirestore(userProfile);
            var uid = user.uid, emailVerified = user.emailVerified;
            _this.store.update({ uid: uid, emailVerified: emailVerified, profile: profile, roles: roles });
        }), map(function (_a) {
            var _b = __read(_a, 3), user = _b[0], userProfile = _b[1], roles = _b[2];
            return user ? [user, _this.formatFromFirestore(userProfile), roles] : null;
        }));
    };
    /**
     * @description Delete user from authentication service and database
     * WARNING This is security sensitive operation
     */
    FireAuthService.prototype.delete = function (options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var user, _a, write, ctx, ref;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this.user];
                    case 1:
                        user = _b.sent();
                        if (!user) {
                            throw new Error('No user connected');
                        }
                        _a = options.write, write = _a === void 0 ? this.db.firestore.batch() : _a, ctx = options.ctx;
                        ref = this.collection.doc(user.uid).ref;
                        write.delete(ref);
                        if (!this.onDelete) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.onDelete({ write: write, ctx: ctx })];
                    case 2:
                        _b.sent();
                        _b.label = 3;
                    case 3:
                        if (!!options.write) return [3 /*break*/, 5];
                        return [4 /*yield*/, write.commit()];
                    case 4:
                        _b.sent();
                        _b.label = 5;
                    case 5: return [2 /*return*/, user.delete()];
                }
            });
        });
    };
    /** Update the current profile of the authenticated user */
    FireAuthService.prototype.update = function (profile, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var user, _a, write, ctx, ref;
            var _this = this;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this.user];
                    case 1:
                        user = _b.sent();
                        if (!user.uid) {
                            throw new Error('No user connected.');
                        }
                        if (!(typeof profile === 'function')) return [3 /*break*/, 2];
                        return [2 /*return*/, this.db.firestore.runTransaction(function (tx) { return __awaiter(_this, void 0, void 0, function () {
                                var ref, snapshot, doc, data;
                                var _a;
                                return __generator(this, function (_b) {
                                    switch (_b.label) {
                                        case 0:
                                            ref = this.collection.doc(user.uid).ref;
                                            return [4 /*yield*/, tx.get(ref)];
                                        case 1:
                                            snapshot = _b.sent();
                                            doc = Object.freeze(__assign(__assign({}, snapshot.data()), (_a = {}, _a[this.idKey] = snapshot.id, _a)));
                                            data = profile(this.formatFromFirestore(doc), tx);
                                            tx.update(ref, data);
                                            if (!this.onUpdate) return [3 /*break*/, 3];
                                            return [4 /*yield*/, this.onUpdate(data, { write: tx, ctx: options.ctx })];
                                        case 2:
                                            _b.sent();
                                            _b.label = 3;
                                        case 3: return [2 /*return*/, tx];
                                    }
                                });
                            }); })];
                    case 2:
                        if (!(typeof profile === 'object')) return [3 /*break*/, 5];
                        _a = options.write, write = _a === void 0 ? this.db.firestore.batch() : _a, ctx = options.ctx;
                        ref = this.collection.doc(user.uid).ref;
                        write.update(ref, this.formatToFirestore(profile));
                        if (!this.onCreate) return [3 /*break*/, 4];
                        return [4 /*yield*/, this.onCreate(profile, { write: write, ctx: ctx })];
                    case 3:
                        _b.sent();
                        _b.label = 4;
                    case 4:
                        // If there is no atomic write provided
                        if (!options.write) {
                            return [2 /*return*/, write.commit()];
                        }
                        _b.label = 5;
                    case 5: return [2 /*return*/];
                }
            });
        });
    };
    /** Create a user based on email and password */
    FireAuthService.prototype.signup = function (email, password, options) {
        if (options === void 0) { options = {}; }
        return __awaiter(this, void 0, void 0, function () {
            var cred, _a, write, ctx, profile, ref;
            return __generator(this, function (_b) {
                switch (_b.label) {
                    case 0: return [4 /*yield*/, this.auth.createUserWithEmailAndPassword(email, password)];
                    case 1:
                        cred = _b.sent();
                        _a = options.write, write = _a === void 0 ? this.db.firestore.batch() : _a, ctx = options.ctx;
                        if (!this.onSignup) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.onSignup(cred, { write: write, ctx: ctx })];
                    case 2:
                        _b.sent();
                        _b.label = 3;
                    case 3: return [4 /*yield*/, this.createProfile(cred.user, ctx)];
                    case 4:
                        profile = _b.sent();
                        ref = this.collection.doc(cred.user.uid).ref;
                        write.set(ref, this.formatToFirestore(profile));
                        if (!this.onCreate) return [3 /*break*/, 6];
                        return [4 /*yield*/, this.onCreate(profile, { write: write })];
                    case 5:
                        _b.sent();
                        _b.label = 6;
                    case 6:
                        if (!!options.write) return [3 /*break*/, 8];
                        return [4 /*yield*/, write.commit()];
                    case 7:
                        _b.sent();
                        _b.label = 8;
                    case 8: return [2 /*return*/, cred];
                }
            });
        });
    };
    FireAuthService.prototype.signin = function (provider, password) {
        return __awaiter(this, void 0, void 0, function () {
            var cred, authProvider, profile, write, ref, err_1;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        this.store.setLoading(true);
                        _a.label = 1;
                    case 1:
                        _a.trys.push([1, 21, , 22]);
                        cred = void 0;
                        if (!!provider) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.auth.signInAnonymously()];
                    case 2:
                        cred = _a.sent();
                        return [3 /*break*/, 11];
                    case 3:
                        if (!(password && typeof provider === 'string')) return [3 /*break*/, 5];
                        return [4 /*yield*/, this.auth.signInWithEmailAndPassword(provider, password)];
                    case 4:
                        cred = _a.sent();
                        return [3 /*break*/, 11];
                    case 5:
                        if (!(typeof provider === 'object')) return [3 /*break*/, 7];
                        return [4 /*yield*/, this.auth.signInWithPopup(provider)];
                    case 6:
                        cred = _a.sent();
                        return [3 /*break*/, 11];
                    case 7:
                        if (!isFireAuthProvider(provider)) return [3 /*break*/, 9];
                        authProvider = getAuthProvider(provider);
                        return [4 /*yield*/, this.auth.signInWithPopup(authProvider)];
                    case 8:
                        cred = _a.sent();
                        return [3 /*break*/, 11];
                    case 9: return [4 /*yield*/, this.auth.signInWithCustomToken(provider)];
                    case 10:
                        cred = _a.sent();
                        _a.label = 11;
                    case 11:
                        if (!cred.additionalUserInfo.isNewUser) return [3 /*break*/, 18];
                        if (!this.onSignup) return [3 /*break*/, 13];
                        return [4 /*yield*/, this.onSignup(cred, {})];
                    case 12:
                        _a.sent();
                        _a.label = 13;
                    case 13: return [4 /*yield*/, this.createProfile(cred.user)];
                    case 14:
                        profile = _a.sent();
                        write = this.db.firestore.batch();
                        ref = this.collection.doc(cred.user.uid).ref;
                        write.set(ref, this.formatToFirestore(profile));
                        if (!this.onCreate) return [3 /*break*/, 16];
                        return [4 /*yield*/, this.onCreate(profile, { write: write })];
                    case 15:
                        _a.sent();
                        _a.label = 16;
                    case 16: return [4 /*yield*/, write.commit()];
                    case 17:
                        _a.sent();
                        return [3 /*break*/, 20];
                    case 18:
                        if (!this.onSignin) return [3 /*break*/, 20];
                        return [4 /*yield*/, this.onSignin(cred)];
                    case 19:
                        _a.sent();
                        _a.label = 20;
                    case 20:
                        this.store.setLoading(false);
                        return [2 /*return*/, cred];
                    case 21:
                        err_1 = _a.sent();
                        this.store.setLoading(false);
                        if (err_1.code === 'auth/operation-not-allowed') {
                            console.warn('You tried to connect with unenabled auth provider. Enable it in Firebase console');
                        }
                        throw err_1;
                    case 22: return [2 /*return*/];
                }
            });
        });
    };
    /** Signs out the current user and clear the store */
    FireAuthService.prototype.signOut = function () {
        return __awaiter(this, void 0, void 0, function () {
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.auth.signOut()];
                    case 1:
                        _a.sent();
                        this.store.update(initialAuthState);
                        if (!this.onSignout) return [3 /*break*/, 3];
                        return [4 /*yield*/, this.onSignout()];
                    case 2:
                        _a.sent();
                        _a.label = 3;
                    case 3: return [2 /*return*/];
                }
            });
        });
    };
    return FireAuthService;
}());

function CollectionGuardConfig(data) {
    return function (constructor) {
        Object.keys(data).forEach(function (key) { return (constructor[key] = data[key]); });
    };
}
var CollectionGuard = /** @class */ (function () {
    function CollectionGuard(service) {
        this.service = service;
        try {
            this.router = inject(Router);
        }
        catch (err) {
            throw new Error('CollectionGuard requires RouterModule to be imported');
        }
    }
    Object.defineProperty(CollectionGuard.prototype, "awaitSync", {
        // Can be override by the extended class
        /** Should the guard wait for connection to Firestore to complete */
        get: function () {
            return this.constructor['awaitSync'] || false;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "queryFn", {
        // Can be override by the extended class
        /** A queryFn to execute for sync */
        get: function () {
            return this.constructor['queryFn'];
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(CollectionGuard.prototype, "redirect", {
        // Can be override by the extended class
        /** The route to redirect to if you sync failed */
        get: function () {
            return this.constructor['redirect'];
        },
        enumerable: true,
        configurable: true
    });
    // Can be override by the extended class
    /** The method to subscribe to while route is active */
    CollectionGuard.prototype.sync = function (next, state) {
        var _a = next.data.queryFn, queryFn = _a === void 0 ? this.queryFn : _a;
        if (this.service instanceof FireAuthService) {
            return this.service.sync();
        }
        else if (this.service instanceof CollectionService) {
            return this.service.syncCollection(queryFn);
        }
    };
    CollectionGuard.prototype.canActivate = function (next, state) {
        var _this = this;
        var _a = next.data, _b = _a.redirect, redirect = _b === void 0 ? this.redirect : _b, _c = _a.awaitSync, awaitSync = _c === void 0 ? this.awaitSync : _c;
        return new Promise(function (res, rej) {
            if (awaitSync) {
                var unsubscribe_1 = new Subject();
                _this.subscription = _this.sync(next, state).pipe(takeUntil(unsubscribe_1)).subscribe({
                    next: function (result) {
                        if (result instanceof UrlTree) {
                            return res(result);
                        }
                        switch (typeof result) {
                            case 'string':
                                unsubscribe_1.next();
                                unsubscribe_1.complete();
                                return res(_this.router.parseUrl(result));
                            case 'boolean':
                                return res(result);
                            default:
                                return res(true);
                        }
                    },
                    error: function (err) {
                        res(_this.router.parseUrl(redirect || ''));
                        throw new Error(err);
                    }
                });
            }
            else {
                _this.subscription = _this.sync(next, state).subscribe();
                res(true);
            }
        });
    };
    CollectionGuard.prototype.canDeactivate = function () {
        if (this.subscription) {
            this.subscription.unsubscribe();
        }
        return true;
    };
    return CollectionGuard;
}());

/** @deprecated Use CollectionService instead */
var CollectionGroupService = /** @class */ (function () {
    function CollectionGroupService(store) {
        this.store = store;
        try {
            this.db = inject(AngularFirestore);
        }
        catch (err) {
            throw new Error('CollectionGroupService requires AngularFirestore.');
        }
    }
    /**
    * Function triggered when getting data from firestore
    * @note should be overrided
    */
    CollectionGroupService.prototype.formatFromFirestore = function (entity) {
        return entity;
    };
    Object.defineProperty(CollectionGroupService.prototype, "idKey", {
        get: function () {
            if (this.store) {
                return this.store.idKey;
            }
        },
        enumerable: true,
        configurable: true
    });
    CollectionGroupService.prototype.syncCollection = function (queryOrOptions, storeOptions) {
        var _this = this;
        if (storeOptions === void 0) { storeOptions = { loading: true }; }
        var query;
        if (typeof queryOrOptions === 'function') {
            query = queryOrOptions;
        }
        else if (typeof queryOrOptions === 'object') {
            storeOptions = queryOrOptions;
        }
        var storeName = getStoreName(this.store, storeOptions);
        // reset has to happen before setLoading, otherwise it will also reset the loading state
        if (storeOptions.reset) {
            resetStore(storeName);
        }
        if (storeOptions.loading) {
            setLoading(storeName, true);
        }
        return this.db.collectionGroup(this.collectionId, query).stateChanges().pipe(withTransaction(function (actions) { return syncStoreFromDocAction(storeName, actions, _this.idKey, function (entity) { return _this.formatFromFirestore(entity); }); }));
    };
    /** Return a snapshot of the collection group */
    CollectionGroupService.prototype.getValue = function (queryGroupFn) {
        return __awaiter(this, void 0, void 0, function () {
            var snapshot;
            var _this = this;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0: return [4 /*yield*/, this.db.collectionGroup(this.collectionId, queryGroupFn).get().toPromise()];
                    case 1:
                        snapshot = _a.sent();
                        return [2 /*return*/, snapshot.docs.map(function (doc) {
                                var entity = doc.data();
                                return _this.formatFromFirestore(entity);
                            })];
                }
            });
        });
    };
    return CollectionGroupService;
}());

function canWrite(write, roles) {
    return roles.some(function (role) { return role === write || role === 'write'; });
}
function canRead(read, roles) {
    return roles.some(function (role) { return role === read || role === 'write'; });
}
function hasRole(role, roles) {
    switch (role) {
        case 'write': return roles.includes('write');
        case 'create': return canWrite('create', roles);
        case 'delete': return canWrite('delete', roles);
        case 'update': return canWrite('update', roles);
        case 'read': return roles.includes('read');
        case 'get': return canRead('get', roles);
        case 'list': return canRead('list', roles);
        default: return false;
    }
}

function shouldCancel(_a) {
    var validate = _a.validate, cancel = _a.cancel;
    return race([validate.pipe(map(function (_) { return false; })), cancel.pipe(map(function (_) { return true; }))]);
}
function waitForCancel(_a) {
    var startWith = _a.startWith, endWith = _a.endWith, shouldValidate = _a.shouldValidate, shouldCancel = _a.shouldCancel;
    return __awaiter(this, void 0, void 0, function () {
        var cancelled;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    startWith();
                    return [4 /*yield*/, race([
                            shouldValidate.pipe(map(function (_) { return false; })),
                            shouldCancel.pipe(map(function (_) { return true; }))
                        ]).toPromise()];
                case 1:
                    cancelled = _b.sent();
                    endWith(cancelled);
                    return [2 /*return*/];
            }
        });
    });
}

/**
 * @description Custom RxJs operator
 * @param redirectTo Route path to redirecto if collection is empty
 */
function redirectIfEmpty(redirectTo) {
    return map(function (actions) { return actions.length === 0 ? redirectTo : true; });
}

function syncWithRouter(routerQuery) {
    var _this = this;
    if (!this['store'].resettable) {
        throw new Error("Store " + this['store'].storeName + " is required to be ressetable for syncWithRouter to work.");
    }
    var pathParams = getPathParams(this.path);
    return routerQuery.selectParams().pipe(
    // Don't trigger change if params needed (and only them) haven't changed
    distinctUntilChanged(function (old, next) {
        var paramsHaveChanged = !!pathParams.find(function (param) { return old[param] !== next[param]; });
        // reset store on every parameter change
        if (paramsHaveChanged) {
            _this['store'].reset();
        }
        return !paramsHaveChanged;
    }), 
    // Need to filter because changes in params comes before canDeactivate
    filter(function (params) { return pathParams.every(function (param) { return !!params[param]; }); }), switchMap(function (params) { return _this.syncCollection({ params: params }); }), share());
}

var queryKeys = ['path', 'queryFn'];
function getSubQueryKeys(query) {
    var keys = Object.keys(query);
    return keys.filter(function (key) { return !queryKeys.includes(key); });
}
function hasSubQueries(query) {
    return getSubQueryKeys(query).length > 0;
}
function getSubQuery(query, parent) {
    if (typeof query !== 'function') {
        return query;
    }
    return query(parent);
}
function isDocPath(path) {
    return path.split('/').length % 2 === 0;
}
/** Transform a path into a collection Query */
function collection(path, queryFn) {
    return { path: path, queryFn: queryFn };
}
/** Transform a path into a doc query */
function doc(path) {
    return { path: path };
}
/** Check if a value is a query */
function isQuery(query) {
    if (typeof query === 'object' && query !== null) {
        return !!query['path'];
    }
    return false;
}

/**
 * Sync the collection
 * @param this Uses this function with bind on a Collection Service
 * @param query The query to trigger
 */
function syncQuery(query) {
    var _this = this;
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this.syncDoc({ path: query })
            : this.syncCollection(query);
    }
    if (Array.isArray(query)) {
        return combineLatest(query.map(function (oneQuery) { return syncQuery.call(_this, oneQuery); }));
    }
    if (!isQuery(query)) {
        throw new Error('Query should be either a path, a Query object or an array of Queries');
    }
    ////////////////
    // SUBQUERIES //
    ////////////////
    /** Listen on Child actions */
    var fromChildAction = function (actions, child) {
        var e_1, _a;
        var idKey = 'id'; // TODO: Improve how to
        var parentId = child.parentId, key = child.key;
        var _loop_1 = function (action) {
            var id = action.payload.doc.id;
            var data = action.payload.doc.data();
            switch (action.type) {
                case 'added': {
                    _this['store'].update(parentId, function (entity) {
                        var _a;
                        return (_a = {},
                            _a[key] = arrayUpsert(entity[key] || [], id, data, idKey),
                            _a);
                    });
                    break;
                }
                case 'removed': {
                    _this['store'].update(parentId, function (entity) {
                        var _a;
                        return (_a = {},
                            _a[key] = arrayRemove(entity[key], id, idKey),
                            _a);
                    });
                    break;
                }
                case 'modified': {
                    _this['store'].update(parentId, function (entity) {
                        var _a;
                        return (_a = {},
                            _a[key] = arrayUpdate(entity[key], id, data, idKey),
                            _a);
                    });
                }
            }
        };
        try {
            for (var actions_1 = __values(actions), actions_1_1 = actions_1.next(); !actions_1_1.done; actions_1_1 = actions_1.next()) {
                var action = actions_1_1.value;
                _loop_1(action);
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (actions_1_1 && !actions_1_1.done && (_a = actions_1.return)) _a.call(actions_1);
            }
            finally { if (e_1) throw e_1.error; }
        }
    };
    /**
     * Stay in sync a subquery of the collection
     * @param subQuery The sub query based on the entity
     * @param child An object that describe the parent
     */
    var syncSubQuery = function (subQuery, child) {
        var _a;
        var parentId = child.parentId, key = child.key;
        // If it's a static value
        // TODO : Check if subQuery is a string ???
        if (!isQuery(subQuery)) {
            var update = _this['store'].update(parentId, (_a = {}, _a[key] = subQuery, _a));
            return of(update);
        }
        if (Array.isArray(subQuery)) {
            var syncQueries = subQuery.map(function (oneQuery) {
                if (isQuery(subQuery)) {
                    var id_1 = getIdAndPath({ path: subQuery.path });
                    return _this['db'].doc(subQuery.path).valueChanges().pipe(tap(function (childDoc) {
                        _this['store'].update(parentId, function (entity) {
                            var _a;
                            return (_a = {},
                                _a[key] = arrayAdd(entity[key], id_1, childDoc),
                                _a);
                        });
                    }));
                }
                return syncSubQuery(oneQuery, child);
            });
            return combineLatest(syncQueries);
        }
        if (typeof subQuery !== 'object') {
            throw new Error('Query should be either a path, a Query object or an array of Queries');
        }
        // Sync subquery
        if (isDocPath(subQuery.path)) {
            return _this['db'].doc(subQuery.path).valueChanges().pipe(tap(function (children) {
                var _a;
                return _this['store'].update(parentId, (_a = {}, _a[key] = children, _a));
            }));
        }
        else {
            return _this['db'].collection(subQuery.path, subQuery.queryFn)
                .stateChanges()
                .pipe(withTransaction(function (actions) { return fromChildAction(actions, child); }));
        }
    };
    /**
     * Get in sync with all the subqueries
     * @param subQueries A map of all the subqueries
     * @param parent The parent entity to attach the children to
     */
    var syncAllSubQueries = function (subQueries, parent) {
        var obs = Object.keys(subQueries)
            .filter(function (key) { return key !== 'path' && key !== 'queryFn'; })
            .map(function (key) {
            var queryLike = getSubQuery(subQueries[key], parent);
            var child = { key: key, parentId: parent[_this.idKey] };
            return syncSubQuery(queryLike, child);
        });
        return combineLatest(obs);
    };
    ////////////////
    // MAIN QUERY //
    ////////////////
    /** Listen on action with child queries */
    var fromActionWithChild = function (actions, mainQuery, subscriptions) {
        var e_2, _a, _b;
        try {
            for (var actions_2 = __values(actions), actions_2_1 = actions_2.next(); !actions_2_1.done; actions_2_1 = actions_2.next()) {
                var action = actions_2_1.value;
                var id = action.payload.doc.id;
                var data = action.payload.doc.data();
                switch (action.type) {
                    case 'added': {
                        var entity = _this.formatFromFirestore(__assign((_b = {}, _b[_this.idKey] = id, _b), data));
                        _this['store'].upsert(id, entity);
                        subscriptions[id] = syncAllSubQueries(mainQuery, entity).subscribe();
                        break;
                    }
                    case 'removed': {
                        _this['store'].remove(id);
                        subscriptions[id].unsubscribe();
                        delete subscriptions[id];
                        break;
                    }
                    case 'modified': {
                        _this['store'].update(id, data);
                    }
                }
            }
        }
        catch (e_2_1) { e_2 = { error: e_2_1 }; }
        finally {
            try {
                if (actions_2_1 && !actions_2_1.done && (_a = actions_2.return)) _a.call(actions_2);
            }
            finally { if (e_2) throw e_2.error; }
        }
    };
    var path = query.path, queryFn = query.queryFn;
    if (isDocPath(path)) {
        var id_2 = getIdAndPath({ path: path }).id;
        var subscription_1;
        return this['db'].doc(path).valueChanges().pipe(tap(function (entity) {
            _this['store'].upsert(id_2, _this.formatFromFirestore(__assign({ id: id_2 }, entity)));
            if (!subscription_1) { // Subscribe only the first time
                subscription_1 = syncAllSubQueries(query, entity).subscribe();
            }
        }), 
        // Stop subscription
        finalize(function () { return subscription_1.unsubscribe(); }));
    }
    else {
        var subscriptions_1 = {};
        return this['db'].collection(path, queryFn)
            .stateChanges()
            .pipe(withTransaction(function (actions) { return fromActionWithChild(actions, query, subscriptions_1); }), 
        // Stop all subscriptions
        finalize(function () { return Object.keys(subscriptions_1).forEach(function (id) {
            subscriptions_1[id].unsubscribe();
            delete subscriptions_1[id];
        }); }));
    }
}

function awaitSyncQuery(query) {
    var _this = this;
    return queryChanges.call(this, query).pipe(tap(function (entities) {
        Array.isArray(entities)
            ? _this['store'].upsertMany(entities)
            : _this['store'].upsert(entities[_this.idKey], entities);
    }));
}
/**
 * Listen on the changes of the documents in the query
 * @param query A query object to listen to
 */
function queryChanges(query) {
    var _this = this;
    return awaitQuery.call(this, query).pipe(map(function (entities) {
        return Array.isArray(entities)
            ? entities.map(function (e) { return _this.formatFromFirestore(e); })
            : _this.formatFromFirestore(entities);
    }));
}
function awaitQuery(query) {
    var _this = this;
    // If single query
    if (typeof query === 'string') {
        return isDocPath(query)
            ? this['db'].doc(query).valueChanges()
            : this['db'].collection(query).valueChanges({ idField: this.idKey });
    }
    if (Array.isArray(query)) {
        return !!query.length
            ? combineLatest(query.map(function (oneQuery) { return awaitSyncQuery.call(_this, oneQuery); }))
            : of(query);
    }
    if (!isQuery(query)) {
        return of(query);
    }
    /**
     * Get the entity of one subquery
     * @param subQuery The subquery function or value
     * @param entity The parent entity
     */
    var syncSubQuery = function (subQueryFn, entity) {
        if (!subQueryFn) {
            return throwError("Query failed");
        }
        if (typeof subQueryFn !== 'function') {
            return of(subQueryFn);
        }
        return awaitQuery.call(_this, subQueryFn(entity));
    };
    /**
     * Get all the Entities of all subqueries
     * @param parentQuery The parent Query
     * @param entity The parent Entity
     */
    var getAllSubQueries = function (parentQuery, entity) {
        if (!entity) {
            return throwError("Nothing found at path : " + parentQuery.path);
        }
        // There is no subqueries return the entity
        if (!hasSubQueries(parentQuery)) {
            return of(entity);
        }
        // Get all subquery keys
        var subQueryKeys = getSubQueryKeys(query);
        // For each key get the subquery
        var subQueries$ = subQueryKeys.map(function (key) {
            return syncSubQuery(parentQuery[key], entity).pipe(tap(function (subentity) { return entity[key] = subentity; }));
        });
        return !!subQueries$.length
            ? combineLatest(subQueries$).pipe(map(function () { return entity; }))
            : of(entity);
    };
    // IF DOCUMENT
    var path = query.path, queryFn = query.queryFn;
    if (isDocPath(path)) {
        var id_1 = getIdAndPath({ path: path }).id;
        return this['db'].doc(path).valueChanges().pipe(switchMap(function (entity) { return getAllSubQueries(query, entity); }), map(function (entity) { return (__assign({ id: id_1 }, entity)); }));
    }
    // IF COLLECTION
    return this['db'].collection(path, queryFn)
        .valueChanges({ idField: this.idKey })
        .pipe(switchMap(function (entities) {
        var entities$ = entities.map(function (entity) { return getAllSubQueries(query, entity); });
        return entities$.length ? combineLatest(entities$) : of([]);
    }));
}

/*
 * Public API Surface of akita-ng-fire
 */

/**
 * Generated bundle index. Do not edit.
 */

export { CollectionConfig, CollectionGroupService, CollectionGuard, CollectionGuardConfig, CollectionService, FireAuthService, authProviders, awaitQuery, awaitSyncQuery, canRead, canWrite, collection, doc, getAuthProvider, getCustomClaims, getIdAndPath, getPathParams, getSubQuery, getSubQueryKeys, hasRole, hasSubQueries, initialAuthState, isDocPath, isFireAuthProvider, isQuery, isTransaction, pathWithParams, queryChanges, queryKeys, redirectIfEmpty, removeStoreEntity, resetStore, setActive, setLoading, shouldCancel, syncQuery, syncStoreFromDocAction, syncStoreFromDocActionSnapshot, syncWithRouter, updateStoreEntity, upsertStoreEntity, waitForCancel };
//# sourceMappingURL=akita-ng-fire.js.map
